import pysam
import subprocess
import pandas as pd
import dnaio


def split_fastq_reads(fastq_path, output_path, trim_b=0, size_l=40, size_r=40, size_m=30):
    """
    Split reads in the fastq file into three parts for remapping.
    Depending on the read length, reads may be
    1) skipped,
    2) split into left and right parts,
    3) split into left, right and middle parts
    left size size_l (name with -l suffix)
    right size size_r (name with -r suffix)
    middle size size_m (name with -m suffix)
    Parameters
    ----------
    fastq_path
    output_path
    trim_b
    size_l
    size_r
    size_m
    Returns
    -------
    """
    trim_b = int(trim_b)
    size_max = max(size_l, size_r)
    with dnaio.open(fastq_path) as f, \
            dnaio.open(output_path, mode='w') as out_f:
        for read in f:
            if trim_b > 0:
                read = read[trim_b:-trim_b]
            read_length = len(read)
            if read_length <= size_max:
                continue
            else:
                # split reads to left and right part, they may have overlap
                left_read = read[:size_l]
                left_read.name += '-l'
                out_f.write(left_read)

                right_read = read[-size_r:]
                right_read.name += '-r'
                out_f.write(right_read)

                # if the middle part is longer enough, we also use it
                if read_length >= (size_l + size_r + size_m):
                    middle_read = read[size_l:-size_r]
                    middle_read.name += '-m'
                    out_f.write(middle_read)
    return


def _output(rfh, tot1, tot2, pre_id, locs):
    c1, c2 = locs[:4].count(''), locs[4:].count('')
    if c1 < 4:
        tot1 += 1
    if c2 < 4:
        tot1 += 1
    if (c1 + c2) < 7:
        tot2 += 1
        rfh.write(pre_id + '\t' + '\t'.join(locs) + '\n')
    return tot1, tot2


def _parse_bam(bam_path, output_path):
    splits = ['1', '1-1', '1-3', '1-2', '2-2', '2-3', '2-1', '2']
    split_dict = {x: i for i, x in enumerate(splits)}
    pre_id = ''
    locs = ['' for _ in range(len(splits))]

    dfh = pysam.AlignmentFile(bam_path, 'rb')
    rfh = open(output_path, 'w')
    tot1, tot2 = 0, 0
    for read in dfh:
        line = str(read).split()
        if '_' in line[0]:
            _id = line[0].split('_')[0]
            split_st = line[0].split('_')[1].split(':')[0]
            if line[0][-2:] == '-l':
                split_st += '-1'
            elif line[0][-2:] == '-r':
                split_st += '-2'
            elif line[0][-2:] == '-m':
                split_st += '-3'
            if read.flag & 16:
                if split_st.split('-')[0] == '1':
                    strand = 0
                else:
                    strand = 1
            else:
                if split_st.split('-')[0] == '1':
                    strand = 1
                else:
                    strand = 0
            if _id != pre_id:
                tot1, tot2 = _output(rfh, tot1, tot2, pre_id, locs)
                pre_id = _id
                locs = ['' for _ in range(len(splits))]
            if strand == 1:
                locs[split_dict[split_st]] = f'1:' \
                                             f'{dfh.get_reference_name(read.reference_id)}:' \
                                             f'{str(read.pos + 1)}'
            if strand == 0:
                locs[split_dict[split_st]] = f'0:' \
                                             f'{dfh.get_reference_name(read.reference_id)}:' \
                                             f'{str(read.pos + len(line[9]))}'
    _output(rfh, tot1, tot2, pre_id, locs)
    rfh.close()

    cmd = f'sort -k2 -u {output_path} -o {output_path}'
    subprocess.run(cmd, shell=True, check=True)
    return output_path


def _parse_split_table(input_path, output_path, chrom_size_path, min_gap=2500):
    output_path = str(output_path)
    if output_path.endswith('.gz'):
        # remove gz first
        output_path = output_path[:-3]

    dfh = open(input_path, 'r')
    rfh = open(output_path, 'w')
    chrom = pd.read_csv(chrom_size_path, sep='\t', header=None, index_col=0).index.tolist()
    chr_dict = {x: i for i, x in enumerate(chrom)}
    tot, cis_long, cis_short, trans = 0, 0, 0, 0

    for line in dfh:
        tot += 1
        frag = line.strip('\n').split('\t')[1:]
        i = 0
        read1 = ''
        for i in range(8):
            if frag[i] != '':
                read1 = frag[i].replace(':', '\t')
                break
        for j in range(7, i, -1):
            if frag[j] != '':
                read2 = frag[j].replace(':', '\t')
                read1_chro, read1_pos = read1.split('\t')[1:3]
                read2_chro, read2_pos = read2.split('\t')[1:3]
                read1_pos, read2_pos = int(read1_pos), int(read2_pos)
                if (read1_chro in chrom) and (read2_chro in chrom):
                    if chr_dict[read1_chro] > chr_dict[read2_chro] or (
                            chr_dict[read1_chro] == chr_dict[read2_chro] and read1_pos > read2_pos):
                        read1, read2 = read2, read1
                    if chr_dict[read1_chro] == chr_dict[read2_chro]:
                        if abs(read1_pos - read2_pos) > min_gap:
                            cis_long += 1
                            rfh.write(read1 + '\t0\t' + read2 + '\t1\n')
                        else:
                            cis_short += 1
                    else:
                        trans += 1
                        rfh.write(read1 + '\t0\t' + read2 + '\t1\n')
                break
    dfh.close()
    rfh.close()
    counts = {'CisShortContact': cis_short,
              'CisLongContact': cis_long,
              'TransContact': trans}
    pd.Series(counts).to_csv(f'{output_path}.counts.txt', header=False)

    # sort the contacts
    subprocess.run(f'sort -k2,2 -k6,6 -k1,1n -k5,5n -k3,3n {output_path} -o {output_path}', shell=True, check=True)

    # compress the contacts, .gz will be added
    subprocess.run(f'gzip {output_path}', shell=True, check=True)
    return output_path


def generate_contacts(bam_path, output_path, chrom_size_path, min_gap=2500, keep_split_table=False):
    split_table_path = f'{output_path}.split.tsv'

    # from bam to eight column split table
    split_table_path = _parse_bam(bam_path=bam_path, output_path=split_table_path)

    # from split table to contacts
    _parse_split_table(input_path=split_table_path,
                       output_path=output_path,
                       chrom_size_path=chrom_size_path,
                       min_gap=min_gap)

    if not keep_split_table:
        subprocess.run(['rm', '-f', split_table_path])
    return



"""
A sbatch wrapper for stampede2
See stampede2 doc:
https://portal.tacc.utexas.edu/user-guides/stampede2#running-sbatch
"""

import pathlib
import re
import shlex
import subprocess
import time
from collections import defaultdict
import pandas as pd
import random
import string

import cemba_data

PACKAGE_DIR = pathlib.Path(cemba_data.__path__[0])

# see stampede2 doc https://portal.tacc.utexas.edu/user-guides/stampede2#running-queues
# name: max_jobs
STAMPEDE2_QUEUES = {
    'development': 1,
    'normal': 50,
    'large': 5,
    'long': 2,
    'flat-quadrant': 5,
    'skx-dev': 1,
    'skx-normal': 20,
    'skx-large': 3
}


def judge_job_success(job_id, retry=3):
    tried = 0
    for i in range(retry):
        try:
            p = subprocess.run(['sacct', '-j', str(job_id), '--format=jobid,exitcode'],
                               check=True, stdout=subprocess.PIPE, encoding='utf8')
            sacct_txt = p.stdout
            pt = re.compile(f'{job_id}\s+0:0')
            if pt.search(sacct_txt):
                return True
            else:
                return False
        except subprocess.CalledProcessError:
            tried += 1
            print(f'sacct error, try again {tried}/{retry}')
            time.sleep(10)
    return False


def get_job_id(sbatch_result):
    """
    parse the sbatch output, check status, and get job id

    Parameters
    ----------
    sbatch_result

    Returns
    -------
    sbatch job_id
    """
    job_id = None
    for line in sbatch_result.split('\n'):
        line = line.strip()
        if line.startswith('-->'):
            # status line
            if line.endswith("OK"):
                continue
            else:
                print(sbatch_result)
                raise ValueError("sbatch output is abnormal, see information above")
        elif line.startswith('Submitted batch job '):
            # job id line
            job_id = line.split(' ')[-1]
        else:
            pass
    if job_id is None:
        print(sbatch_result)
        raise ValueError('Can not get job id from sbatch output, see information above')
    return job_id


def submit_sbatch(job_script_path):
    """
    submit sbatch job and return job id

    Parameters
    ----------
    job_script_path

    Returns
    -------

    """
    try:
        p = subprocess.run(['sbatch', str(job_script_path)],
                           check=True,
                           encoding='utf8',
                           stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE)
        job_id = get_job_id(p.stdout)
        print(f'Submit job script: {job_script_path}. Job ID is {job_id}.')
    except subprocess.CalledProcessError as e:
        print('sbatch STDOUT')
        print(e.stdout)
        print('sbatch STDERR')
        print(e.stderr)
        print('sbatch submission failed, see information above.')
        raise e
    return job_id


def squeue(partition):
    """
    check current running job

    Returns
    -------
    squeue results in a pd.DataFrame
    """
    user_name = subprocess.run(['whoami'],
                               check=True,
                               encoding='utf8',
                               stdout=subprocess.PIPE).stdout.strip()
    for i in range(3):
        try:
            squeue_result = subprocess.run(['squeue', '-u', user_name],
                                           check=True,
                                           encoding='utf8',
                                           stdout=subprocess.PIPE).stdout
            break
        except subprocess.CalledProcessError:
            print(f'Squeue got an error, waiting 60s and trying again {i + 1}/3')
            time.sleep(60)
            continue
    else:
        raise SystemError('Squeue command failed')
    print('Current squeue output:')
    print(squeue_result, end='\n')

    records = []
    col_names = []
    col_end_pos = []
    for i, line in enumerate(squeue_result.rstrip().split('\n')):
        if i == 0:
            sep_pattern = re.compile(r' +')
            col_names = sep_pattern.split(line.strip())
            col_end_pos = [0] + [line.index(col_name) + len(col_name) for col_name in col_names]
        if line == '':
            continue
        record = []
        for j in range(len(col_end_pos) - 1):
            if j != len(col_names) - 1:
                col_data = line[col_end_pos[j]:col_end_pos[j + 1]]
            else:
                # for last column, take all the rest chr
                col_data = line[col_end_pos[j]:]
            record.append(col_data.strip())
        records.append(record)
    squeue_df = pd.DataFrame(records[1:],
                             columns=records[0]).set_index('JOBID')
    total_job = squeue_df.shape[0]
    try:
        squeue_df = squeue_df[squeue_df['PARTITION'].str.lower() == partition.lower()].copy()
        return squeue_df, total_job
    except KeyError:
        print(squeue_df)
        return squeue_df, total_job


def make_sbatch_script_files(commands, sbatch_dir, name_prefix, queue, time_str, email, email_type):
    """See stampede2 doc: https://portal.tacc.utexas.edu/user-guides/stampede2#running-sbatch"""
    with open(PACKAGE_DIR / 'files/sbatch_template.txt') as f:
        sbatch_template = f.read()
    sbatch_dir = pathlib.Path(sbatch_dir)

    if email is not None:
        email_str = f'#SBATCH --mail-user={email}'
        email_type_str = f'#SBATCH --mail-type={email_type}'
    else:
        email_str = ''
        email_type_str = ''

    env_dir_random = ''.join(random.choices(string.ascii_uppercase + string.digits, k=6))

    queue_job_path_list = []
    for i, command in enumerate(commands):
        job_name = f'{name_prefix}_{i}'
        sbatch_script = sbatch_template.format(
            job_name=job_name,
            queue=queue,
            time_str=time_str,
            email_str=email_str,
            email_type_str=email_type_str,
            command=command,
            log_dir=sbatch_dir,
            env_dir_random=env_dir_random
        )
        job_script_path = sbatch_dir / f'{job_name}.sh'
        with open(job_script_path, 'w') as f:
            f.write(sbatch_script)
        queue_job_path_list.append(job_script_path)
    return queue_job_path_list


def sacct(jobs):
    sep_pattern = re.compile(r' +')
    sacct_cmd = f'sacct -j {",".join(jobs)} ' \
                f'--format=jobid,jobname,partition,alloccpus,elapsed,state,exitcode'
    sacct_result = subprocess.run(shlex.split(sacct_cmd),
                                  check=True,
                                  encoding='utf8',
                                  stdout=subprocess.PIPE).stdout
    lines = []
    header = ''
    col_starts = []
    for i, line in enumerate(sacct_result.rstrip('\n').split('\n')):
        if i == 0:
            header = line
        elif i == 1:
            # the second row indicates col width, use it to determine col_starts
            col_width = [len(s) for s in sep_pattern.split(line)]
            cur_pos = 0
            col_starts = [0]
            for length in col_width:
                cur_pos += (length + 1)
                col_starts.append(cur_pos)
        else:
            lines.append(line)

    columns = [header[col_starts[i]:col_starts[i + 1]].strip() for i in range(len(col_starts) - 1)]
    data = []
    for line in lines:
        ll = [line[col_starts[i]:col_starts[i + 1]].strip() for i in range(len(col_starts) - 1)]
        data.append(ll)
    sacct_data = pd.DataFrame(data, columns=columns).set_index('JobID')
    sacct_data = sacct_data[~sacct_data.index.str.endswith('bat+')].copy()
    sacct_data['Success'] = sacct_data['ExitCode'] == '0:0'
    return sacct_data


def sbatch_submitter(project_name, command_file_path, working_dir, time_str, queue='skx-normal',
                     email=None, email_type='fail', max_jobs=None, dry_run=False, retry=2):
    # read commands
    with open(command_file_path) as f:
        # I always assume the command is ordered with descending priority.
        # But sbatch will submit last job first (list.pop), so reverse the order here.
        commands = [line.rstrip('\n') for line in f if not line.startswith('#')][::-1]

    # set name
    project_name = project_name.replace(' ', '_')

    # check queue
    queue = queue.lower()
    if queue not in STAMPEDE2_QUEUES:
        raise KeyError(f'queue name {queue} not found in STAMPEDE2_QUEUES, '
                       f'available queues are {list(STAMPEDE2_QUEUES.keys())}')

    # set max_jobs
    _max_jobs = STAMPEDE2_QUEUES[queue]
    if max_jobs is None:
        max_jobs = _max_jobs
    else:
        max_jobs = min(max_jobs, _max_jobs)
    print(f'Max concurrent sbatch jobs {max_jobs}, stampede2 allows {STAMPEDE2_QUEUES[queue]}.')

    # make sbatch_dir
    sbatch_dir = pathlib.Path(working_dir) / f'{project_name}_sbatch'
    sbatch_dir.mkdir(exist_ok=True, parents=True)

    # check if sacct file exists, which could from previous submission.
    # I only keep successful items, and skip them in this submission.
    sacct_path = sbatch_dir / 'sacct.csv.gz'
    previous_sacct_df_success = None
    successful_script_paths = set()
    if sacct_path.exists():
        print('Found previous submission records, successful jobs will not be submit again.')
        previous_sacct_df = pd.read_csv(sacct_path, index_col=0)
        previous_sacct_df_success = previous_sacct_df[previous_sacct_df['Success']]
        successful_script_paths = set(previous_sacct_df_success['ScriptPath'].astype(str).tolist())
        print(f'Successful script paths: {", ".join(successful_script_paths)}')

    # create job script files
    queue_job_path_list = make_sbatch_script_files(
        commands=commands,
        sbatch_dir=sbatch_dir,
        name_prefix=project_name,
        queue=queue,
        time_str=time_str,
        email=email,
        email_type=email_type)

    # prepare submission
    running_job_id_set = set()  # sbatch_id
    finished_job_id_set = set()  # job_id
    job_id_to_script_path = {}

    # start submission
    sleepy = 30
    flag_path = sbatch_dir / 'RUNNING_SIGNAL'
    if flag_path.exists():
        raise FileExistsError(f'Running signal exists {flag_path}. '
                              f'Make sure you do not have sbatch submitter running and (if so) delete that flag file.')
    with open(flag_path, 'w') as f:
        f.write('')
    script_path_to_tried_times = defaultdict(int)
    if not dry_run:
        squeue_fail = 0
        while (len(queue_job_path_list) != 0) or (len(running_job_id_set) != 0):
            if not flag_path.exists():
                # break if flag missing
                break
            # squeue and update running job status
            try:
                squeue_df, total_job = squeue(partition=queue)
                squeue_fail = 0
            except Exception as e:
                print('Squeue parser raised an error, will retry after 150s.')
                squeue_fail += 1
                if squeue_fail > 10:
                    raise e
                time.sleep(150)
                continue
            # queue limit and total job limit both apply
            remaining_slots = min((max_jobs - squeue_df.shape[0], 50 - total_job))
            # the max_jobs is apply to user level, not to the current submitter level
            if remaining_slots > 0:
                # things are getting done, weak up
                sleepy = 30

                # check running jobs
                new_running_job_id_set = set()
                for job_id in running_job_id_set:
                    if job_id not in squeue_df.index:
                        # running job finished
                        if judge_job_success(job_id):
                            # job succeed
                            finished_job_id_set.add(job_id)
                        else:
                            # job failed
                            script_path = job_id_to_script_path[job_id]
                            script_path_to_tried_times[script_path] += 1
                            if script_path_to_tried_times[script_path] <= retry:
                                print(f'Job {job_id} failed, retry {script_path} '
                                      f'{script_path_to_tried_times[script_path]}/{retry}.')
                                queue_job_path_list.append(script_path)
                            else:
                                # add the last job_id into finished
                                finished_job_id_set.add(job_id)
                                print(f'{script_path} failed after {retry + 1} attempts.')
                        # status will be judged in the end
                    else:
                        # still running
                        new_running_job_id_set.add(job_id)
                        pass
                running_job_id_set = new_running_job_id_set
                print(f'{len(running_job_id_set)} running job IDs: {", ".join(running_job_id_set)}')

                # submit new jobs
                while (remaining_slots > 0) and (len(queue_job_path_list) > 0):
                    print(f'Remaining slots: {remaining_slots}')
                    script_path = queue_job_path_list.pop()
                    # skip if job already submitted and are successful before
                    if str(script_path) in successful_script_paths:
                        print(f'Already successful in previous submission: {script_path}')
                        continue
                    job_id = submit_sbatch(script_path)
                    running_job_id_set.add(job_id)
                    job_id_to_script_path[job_id] = script_path
                    remaining_slots -= 1

            # sleep
            sleepy += 30
            sleepy = min(300, sleepy)
            time.sleep(sleepy)

    # only check status if something has finished
    if len(finished_job_id_set) > 0:
        # check status
        chunk_size = 50
        stats = []
        finished_job_ids = list(finished_job_id_set)
        for i in range(0, len(finished_job_ids), chunk_size):
            job_chunk = finished_job_ids[i: i + chunk_size]
            stats.append(sacct(job_chunk))
        sacct_df = pd.concat(stats)
        sacct_df['ScriptPath'] = sacct_df.index.map(job_id_to_script_path)

        # add previous sacct records here
        if previous_sacct_df_success is not None:
            sacct_df = pd.concat([previous_sacct_df_success, sacct_df], sort=True)
        sacct_df.to_csv(sacct_path)

        # print time elapsed
        times = pd.to_timedelta(sacct_df[sacct_df['Success']]['Elapsed'])
        print('For all the successful jobs')
        print('The minimum job elapsed:', times.min())
        print('The average job elapsed:', times.mean())
        print('The maximum job elapsed:', times.max())

        # print success
        print(f"{sacct_df['Success'].sum()} / {sacct_df.shape[0]} succeeded")

    # delete flag
    subprocess.run(shlex.split(f'rm -f {flag_path}'), check=True)
    return



"""
When adding new function:
1. add a func_register_subparser function to register the subparser
2. add a condition in main func about this new func name, import the real func as func in main
"""

import argparse
import inspect
import logging
import sys

import cemba_data
from .__main__ import setup_logging

log = logging.getLogger()

DESCRIPTION = """
yap-internal is used for automation, not intend to be used by end user. 
Use yap instead. 
"""

EPILOG = ''


def select_dna_reads_internal_subparser(subparser):
    parser = subparser.add_parser('select-dna-reads',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Select DNA reads, see cemba_data.mapping.mct_bismark_bam_filter")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--input_bam",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--output_bam",
        type=str,
        required=True
    )

    parser.add_argument(
        "--mc_rate_max_threshold",
        type=float,
        default=0.5,
    )

    parser.add_argument(
        "--cov_min_threshold",
        type=int,
        default=5
    )
    return


def select_rna_reads_internal_subparser(subparser):
    parser = subparser.add_parser('select-rna-reads',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Select RNA reads, see cemba_data.mapping.mct_star_bam_filter")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--input_bam",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--output_bam",
        type=str,
        required=True
    )

    parser.add_argument(
        "--mc_rate_min_threshold",
        type=float,
        default=0.9,
    )

    parser.add_argument(
        "--cov_min_threshold",
        type=int,
        default=5
    )
    return


def featurecount_internal_subparser(subparser):
    parser = subparser.add_parser('featurecount',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Batch FeatureCount wrapper")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--bam_table",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--out_prefix",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--gtf_path",
        type=str,
        required=True
    )

    parser.add_argument(
        "--count_type",
        type=str,
        default='gene',
    )

    parser.add_argument(
        "--id_type",
        type=str,
        default='gene_id',
    )

    parser.add_argument(
        "--cpu",
        type=int,
        default=2
    )

    parser.add_argument(
        "--chunksize",
        type=int,
        default=50
    )
    return


def atac_bulk_pipeline_internal_subparser(subparser):
    parser = subparser.add_parser('atac-bulk-pipeline',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Simple Wrapper. "
                                       "Make cluster level merged fragment bed, bigwig "
                                       "and MACS2 peaks from ATAC clustering results.")

    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--cell_group_path",
        type=str,
        required=True,
        help="first row is header, names of cluster columns will be used as output names; "
             "Column 0 is sample name; "
             "Column 1 is cell barcode; "
             "Column 2, ..., n are level of cell group/clusters;"
    )

    parser_req.add_argument(
        "--output_dir_path",
        type=str,
        required=True,
        help="Output directory, each cluster col will be a sub-dir"
    )

    parser_req.add_argument(
        "--sample_snap_path",
        type=str,
        required=True,
        help="no header; "
             "Column 0 is sample name; "
             "Column 1 is sample SNAP file path;"
    )

    parser_req.add_argument(
        "--chrom_size_path",
        type=str,
        required=True,
        help="chromosome size file path"
    )

    parser_req.add_argument(
        "--species",
        type=str,
        required=True,
        help="hs or mm"
    )

    parser.add_argument(
        "--cpu",
        type=int,
        default=1,
        help="Number of cpu to parallel"
    )


def mapping_summary_internal_subparser(subparser):
    parser = subparser.add_parser('summary',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="mapping summary CLI for internal ues")

    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--output_dir",
        type=str,
        required=True,
        help="output_dir of each UID"
    )


def split_read_internal_subparser(subparser):
    parser = subparser.add_parser('m3c-split-reads',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Split unmapped reads for remap in snm3C-seq data")

    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--fastq_path",
        type=str,
        required=True,
        help="Input fastq path"
    )

    parser_req.add_argument(
        "--output_path",
        type=str,
        required=True,
        help="Output fastq path"
    )

    parser_req.add_argument(
        "--trim_b",
        type=int,
        default=0,
        help="Whether trim on both ends before splitting reads. Default is not trim."
    )

    parser_req.add_argument(
        "--size_l",
        type=int,
        default=40,
        help="Size to slice on the left part"
    )

    parser_req.add_argument(
        "--size_r",
        type=int,
        default=40,
        help="Size to slice on the right part"
    )

    parser_req.add_argument(
        "--size_m",
        type=int,
        default=30,
        help="Minimum size of the middle part"
    )


def generate_contacts_internal_subparser(subparser):
    parser = subparser.add_parser('generate-contacts',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="parse ")

    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--bam_path",
        type=str,
        required=True,
        help="Input bam path"
    )

    parser_req.add_argument(
        "--output_path",
        type=str,
        required=True,
        help="Output contact file path"
    )

    parser_req.add_argument(
        "--chrom_size_path",
        type=str,
        required=True,
        help="Chrom size path"
    )

    parser_req.add_argument(
        "--min_gap",
        type=int,
        default=2500,
        help="Minimum gap distance to be considered as contact"
    )

    parser.add_argument(
        "--keep_split_table",
        dest='keep_split_table',
        action='store_true',
        help='Keep the reads split table before contacts table?'
    )
    parser.set_defaults(keep_split_table=False)


def dss_two_internal_subparser(subparser):
    parser = subparser.add_parser('dss-two',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Run DSS two-group DMR")

    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--allc_table_path",
        type=str,
        required=True,
        help="Three columns separated by tab: 1) allc path; 2) sample id 3) group id"
    )

    parser_req.add_argument(
        "--output_dir",
        type=str,
        required=True,
        help="Output directory"
    )

    parser_req.add_argument(
        "--study_name",
        type=str,
        required=True,
        help="Name of the study"
    )

    parser_req.add_argument(
        "--chrom_sizes_path",
        type=str,
        required=True,
        help="Path to genome chrom size file"
    )

    parser.add_argument(
        "--chroms",
        type=str,
        nargs='+',
        default=None,
        help="Chromosomes to consider."
    )

    parser.add_argument("--p_threshold", type=float, default=0.001,
                        help="FDR threshold to select sig DML")
    parser.add_argument("--min_cg", type=int, required=True, default=1,
                        help="Minimum CpGs for a DMR")
    parser.add_argument("--min_len", type=int, required=True, default=1,
                        help="Minimum length for a DMR")
    parser.add_argument("--sig_ratio", type=float, required=True, default=0.5,
                        help="Minimum ratio of CpGs that are significant in a DMR.")
    parser.add_argument("--delta", type=float, required=True, default=0.1,
                        help="Methylation delta that considered to be informative.")
    parser.add_argument("--cpu", type=int, required=True, default=10,
                        help="Number of CPUs to use")
    parser.add_argument("--chunk_size", type=int, required=False, default=50000000,
                        help="chunk size to parallel jobs")
    parser.add_argument(
        "--not_smoothing",
        dest='smoothing',
        action='store_false',
        help='Do not preform smoothing (will perform smoothing by default).'
    )
    parser.set_defaults(smoothing=True)
    parser.add_argument(
        "--save_dml",
        dest='save_dml',
        action='store_true',
        help='Save all the DML files (will delete DML table by default).'
    )
    parser.set_defaults(save_dml=False)
    return

def dss_multi_internal_subparser(subparser):
    parser = subparser.add_parser('dss-multi',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Run DSS two-group DMR")

    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--allc_table_path",
        type=str,
        required=True,
        help="Three columns separated by tab: 1) allc path; 2) sample id 3) group id"
    )

    parser_req.add_argument(
        "--output_dir",
        type=str,
        required=True,
        help="Output directory"
    )

    parser_req.add_argument(
        "--study_name",
        type=str,
        required=True,
        help="Name of the study"
    )

    parser_req.add_argument(
        "--chrom_sizes_path",
        type=str,
        required=True,
        help="Path to genome chrom size file"
    )

    parser.add_argument(
        "--chroms",
        type=str,
        nargs='+',
        default=None,
        help="Chromosomes to consider."
    )

    parser.add_argument("--p_threshold", type=float, default=0.001,
                        help="FDR threshold to select sig DML")
    parser.add_argument("--min_cg", type=int, required=True, default=1,
                        help="Minimum CpGs for a DMR")
    parser.add_argument("--min_len", type=int, required=True, default=1,
                        help="Minimum length for a DMR")
    parser.add_argument("--sig_ratio", type=float, required=True, default=0.5,
                        help="Minimum ratio of CpGs that are significant in a DMR.")
    parser.add_argument("--cpu", type=int, required=True, default=10,
                        help="Number of CPUs to use")
    parser.add_argument("--chunk_size", type=int, required=False, default=50000000,
                        help="chunk size to parallel jobs")
    parser.add_argument(
        "--no_smooth",
        dest='smoothing',
        action='store_false',
        help='Do not preform smoothing (will perform smoothing by default).'
    )
    parser.set_defaults(smoothing=True)
    parser.add_argument(
        "--save_dml",
        dest='save_dml',
        action='store_true',
        help='Save all the DML files (will delete DML table by default).'
    )
    parser.set_defaults(save_dml=False)
    return


def dmrseq_internal_subparser(subparser):
    parser = subparser.add_parser('dmrseq',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Run DMRseq")

    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--allc_table_path",
        type=str,
        required=True,
        help="Three columns separated by tab: 1) allc path; 2) sample id 3) group id"
    )

    parser_req.add_argument(
        "--output_dir",
        type=str,
        required=True,
        help="Output directory"
    )

    parser_req.add_argument(
        "--study_name",
        type=str,
        required=True,
        help="Name of the study"
    )

    parser_req.add_argument(
        "--chrom_sizes_path",
        type=str,
        required=True,
        help="Path to genome chrom size file"
    )

    parser.add_argument(
        "--chroms",
        type=str,
        nargs='+',
        default=None,
        help="Chromosomes to consider."
    )

    parser.add_argument("--test_covariate", type=str, default='group',
                        help="Test covariate name in the allc path table")
    parser.add_argument("--match_covariate", type=str, required=False, default=None, nargs='+',
                        help="Matched covariate name in the allc path table (will be adjusted during permutation.")
    parser.add_argument("--adjust_covariate", type=str, required=False, default=None, nargs='+',
                        help="adjust covariate name in the allc path table (will be adjusted in the model.")
    parser.add_argument("--cutoff", type=float, required=False, default=0.1,
                        help="Methylation delta that considered to be informative.")
    parser.add_argument("--min_num_region", type=int, required=False, default=3,
                        help="Minimum number of CpGs that are significant in a DMR, min is 3.")
    parser.add_argument("--bp_span", type=int, required=False, default=1000,
                        help="")
    parser.add_argument("--min_in_span", type=int, required=False, default=30,
                        help="")
    parser.add_argument("--max_gap_smooth", type=int, required=False, default=2500,
                        help="")
    parser.add_argument("--max_gap", type=int, required=False, default=1000,
                        help="")
    parser.add_argument("--max_perms", type=int, required=False, default=10,
                        help="")
    parser.add_argument("--stat", type=str, required=False, default='stat',
                        help="")
    parser.add_argument("--chrs_per_chunk", type=int, required=False, default=1,
                        help="")
    parser.add_argument("--block_size", type=int, required=False, default=5000,
                        help="")
    parser.add_argument("--cpu", type=int, required=False, default=4,
                        help="")
    parser.add_argument("--template_path", type=str, required=False, default='default',
                        help="")
    parser.add_argument("--chunk_size", type=int, required=False, default=50000000,
                        help="chunk size to parallel jobs")
    parser.add_argument(
        "--no_smooth",
        dest='smooth',
        action='store_false',
        help='Do not preform smoothing (will perform smoothing by default).'
    )
    parser.set_defaults(smooth=True)
    parser.add_argument(
        "--block",
        dest='block',
        action='store_true',
        help='Save all the DML files (will delete DML table by default).'
    )
    parser.set_defaults(block=False)
    return


def internal_main():
    parser = argparse.ArgumentParser(description=DESCRIPTION,
                                     epilog=EPILOG,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    subparsers = parser.add_subparsers(
        title="functions",
        dest="command",
        metavar=""
    )

    # add subparsers
    current_module = sys.modules[__name__]
    # get all functions in parser
    for name, register_subparser_func in inspect.getmembers(current_module, inspect.isfunction):
        if 'internal_subparser' in name:
            register_subparser_func(subparsers)

    # initiate
    args = None
    if len(sys.argv) > 1:
        # print out version
        if sys.argv[1] in ['-v', '--version']:
            print(cemba_data.__version__)
            exit()
        else:
            args = parser.parse_args()
    else:
        # print out help
        parser.parse_args(["-h"])
        exit()

    # set up logging
    if not logging.root.handlers:
        setup_logging(stdout=True,
                      quiet=False)

    # execute command
    args_vars = vars(args)
    cur_command = args_vars.pop('command')
    # Do real import here:
    if cur_command == 'select-dna-reads':
        from .mapping.mct import select_dna_reads as func
    elif cur_command == 'select-rna-reads':
        from .mapping.mct import select_rna_reads as func
    elif cur_command == 'summary':
        from .mapping.stats import mapping_stats as func
    elif cur_command == 'm3c-split-reads':
        from .mapping.m3c import split_fastq_reads as func
    elif cur_command == 'generate-contacts':
        from .mapping.m3c import generate_contacts as func
    elif cur_command == 'dss-two':
        from .dmr.dss import run_dss_two_group as func
    elif cur_command == 'dss-multi':
        from .dmr.dss import run_dss_multi_group as func
    elif cur_command == 'dmrseq':
        from .dmr.dmrseq import run_dmrseq as func
    else:
        log.debug(f'{cur_command} not Known, check the main function if else part')
        parser.parse_args(["-h"])
        return

    # run the command
    log.info(f"# Executing {cur_command}...")
    func(**args_vars)
    log.info(f"# {cur_command} finished.")
    return



import pathlib
import pandas as pd
import subprocess
import re


def execute_command(cmd):
    print(cmd)
    try:
        subprocess.run(cmd,
                       stderr=subprocess.PIPE,
                       encoding='utf8',
                       stdout=subprocess.PIPE,
                       check=True,
                       shell=True)
    except subprocess.CalledProcessError as e:
        print(e.stderr)
        raise e
    return


def modify_snakefile_for_sbatch(path, project_name):
    """Change the input_scool line in snakefile to remote scratch path"""
    with open(path) as f:
        lines = f.readlines()

    with open(path, 'w') as f:
        for line in lines:
            if line.startswith('input_scool'):
                original_path = line.strip().strip("'").split(' = ')[1]
                original_path = pathlib.Path(original_path)
                new_path = f"f'{{scratch_path}}/{project_name}/scool/raw/{original_path.name}'"
                new_line = f'''
import subprocess
scratch_path = subprocess.run('echo $SCRATCH', 
                              shell=True,  
                              check=True,  
                              stdout=subprocess.PIPE,  
                              encoding='utf8').stdout.strip()
input_scool = {new_path}
'''
                f.write(new_line)
            else:
                f.write(line)
    return


def prepare_impute_dir(output_dir,
                       chrom_size_path,
                       scheduler='sbatch',
                       scool_cpu=5,
                       cpu_per_job=10,
                       batch_size=100,
                       total_cpu=100,
                       min_contacts_per_cell=50000,
                       sbatch_time_str='2:00:00',
                       min_cutoff='1e-5',
                       skip_scool_prep=False):
    # set up directory structure under output_dir
    output_dir = pathlib.Path(output_dir).absolute()
    project_name = output_dir.name
    scool_dir = output_dir / 'scool'
    snakemake_dir = scool_dir / 'snakemake'
    snakemake_dir.mkdir(exist_ok=True, parents=True)
    raw_dir = scool_dir / 'raw'
    raw_dir.mkdir(exist_ok=True)
    impute_dir = scool_dir / 'impute'
    impute_dir.mkdir(exist_ok=True)
    dataset_dir = scool_dir / 'dataset'
    dataset_dir.mkdir(exist_ok=True)

    # make contacts
    contact_paths = pd.Series({path.name.split('.')[0]: str(path)
                               for path in output_dir.glob('*/hic/*contact.tsv.gz')})

    # filter by minimum contacts cutoff
    cell_meta = pd.read_csv(f'{output_dir}/stats/MappingSummary.csv.gz', index_col=0)
    use_cells = cell_meta[cell_meta['CisLongContact'] > min_contacts_per_cell].index
    print(f'{cell_meta.shape[0]} cells in total, '
          f'{use_cells.size} cells having contacts > {min_contacts_per_cell} cutoff.')
    contact_paths = contact_paths.loc[use_cells].copy()

    contact_table_path = scool_dir / 'contacts_table.tsv'
    contact_paths.to_csv(contact_table_path, sep='\t', header=None)

    # target scool file paths
    scool_path_10k = raw_dir / f'{project_name}.10K.scool'
    scool_path_25k = raw_dir / f'{project_name}.25K.scool'
    scool_path_100k = raw_dir / f'{project_name}.100K.scool'

    # generate scool command
    if skip_scool_prep:
        if not scool_path_10k.exists():
            raise FileNotFoundError(f'skip_scool_prep is True, but {scool_path_10k} not found.')
        if not scool_path_25k.exists():
            raise FileNotFoundError(f'skip_scool_prep is True, but {scool_path_25k} not found.')
        if not scool_path_100k.exists():
            raise FileNotFoundError(f'skip_scool_prep is True, but {scool_path_100k} not found.')
    else:
        scool_cmd = f'hicluster generate-scool --contacts_table {contact_table_path} ' \
                f'--output_prefix {raw_dir / project_name} ' \
                f'--chrom_size_path {chrom_size_path} ' \
                f'--resolutions 10000 25000 100000 ' \
                f'--min_pos_dist 2500 --cpu {scool_cpu}'
        print('Generating raw scool files for 10Kb, 25Kb and 100Kb resolutions.')
        execute_command(scool_cmd)

    # prepare imputation commands
    # Prepare 10K impute
    this_out_dir = pathlib.Path(f'{impute_dir}/10K/')
    this_out_dir.mkdir(exist_ok=True)
    impute_10k_cmd = f'hicluster imputation ' \
                     f'--input_scool {scool_path_10k} ' \
                     f'--output_dir {this_out_dir} ' \
                     f'--chrom_size_path {chrom_size_path} ' \
                     f'--output_dist 5050000 ' \
                     f'--window_size 30000000 ' \
                     f'--step_size 10000000 ' \
                     f'--resolution 10000 ' \
                     f'--batch_size {batch_size} ' \
                     f'--pad 2 --std 1 --rp 0.5 --tol 0.01 ' \
                     f'--min_cutoff {min_cutoff} ' \
                     f'--cpu_per_job {cpu_per_job}'
    print('Prepare imputation snakefiles for 10Kb resolution.')
    execute_command(impute_10k_cmd)

    # 25K
    this_out_dir = pathlib.Path(f'{impute_dir}/25K/')
    this_out_dir.mkdir(exist_ok=True)
    impute_25k_cmd = f'hicluster imputation ' \
                     f'--input_scool {scool_path_25k} ' \
                     f'--output_dir {this_out_dir} ' \
                     f'--chrom_size_path {chrom_size_path} ' \
                     f'--output_dist 10050000 ' \
                     f'--window_size 500000000 ' \
                     f'--step_size 500000000 ' \
                     f'--resolution 25000 ' \
                     f'--batch_size {batch_size} ' \
                     f'--pad 2 --std 1 --rp 0.5 --tol 0.01 ' \
                     f'--min_cutoff {min_cutoff} ' \
                     f'--cpu_per_job {cpu_per_job}'
    print('Prepare imputation snakefiles for 25Kb resolution.')
    execute_command(impute_25k_cmd)

    # 100K
    this_out_dir = pathlib.Path(f'{impute_dir}/100K/')
    this_out_dir.mkdir(exist_ok=True)
    impute_100k_cmd = f'hicluster imputation ' \
                      f'--input_scool {scool_path_100k} ' \
                      f'--output_dir {this_out_dir} ' \
                      f'--chrom_size_path {chrom_size_path} ' \
                      f'--output_dist 500000000 ' \
                      f'--window_size 500000000 ' \
                      f'--step_size 500000000 ' \
                      f'--resolution 100000 ' \
                      f'--batch_size {batch_size} ' \
                      f'--pad 1 --std 1 --rp 0.5 --tol 0.01 ' \
                      f'--min_cutoff {min_cutoff} ' \
                      f'--cpu_per_job {cpu_per_job}'
    print('Prepare imputation snakefiles for 100Kb resolution.')
    execute_command(impute_100k_cmd)

    if scheduler == 'qsub':
        # prepare qsub
        qsub_dir = snakemake_dir / 'qsub'
        qsub_dir.mkdir(exist_ok=True)
        subprocess.run(f'cat {impute_dir}/*/snakemake_cmd.txt > {qsub_dir}/snakemake_cmd.txt',
                       shell=True, check=True)
        qsub_str = f'yap qsub --command_file_path {qsub_dir}/snakemake_cmd.txt ' \
                   f'--working_dir {qsub_dir} --project_name y{project_name} ' \
                   f'--total_cpu {total_cpu} --qsub_global_parms "-pe smp={cpu_per_job};-l h_vmem=5G"'
        with open(qsub_dir / 'qsub.sh', 'w') as f:
            f.write(qsub_str)
    elif scheduler == 'sbatch':
        # an important notice
        print(f'\n\n\nYou used sbatch as scheduler, remember to '
              f'1) provide the remote path of chrom_size_path; '
              f'2) transfer the {project_name}/scool directory to $SCRATCH/{project_name}/scool. '
              f'Otherwise the snakemake will raise error.')

        # prepare sbatch
        local_sbatch_dir = snakemake_dir / 'sbatch'
        remote_sbatch_dir = f'$SCRATCH/{project_name}/scool/snakemake/sbatch'
        local_sbatch_dir.mkdir(exist_ok=True)
        snake_command_files = impute_dir.glob('*/snakemake_cmd.txt')
        # concat all the snakemake commands and swap dir path
        with open(f'{local_sbatch_dir}/snakemake_cmd.txt', 'w') as out_f:
            for path in snake_command_files:
                with open(path) as f:
                    for line in f:
                        regex_txt = f'(?<= )\S+?(?=/{project_name}/scool/impute/)'
                        p = re.compile(regex_txt)
                        line = p.sub('$SCRATCH', line)
                        out_f.write(line)
        sbatch_str = f'yap sbatch --project_name {project_name} ' \
                     f'--command_file_path {remote_sbatch_dir}/snakemake_cmd.txt ' \
                     f'--working_dir {remote_sbatch_dir} --time_str {sbatch_time_str}'
        with open(local_sbatch_dir / 'sbatch.sh', 'w') as f:
            f.write(sbatch_str)

        # finally the path in snakefile need to be changed, go though each snakefile
        for path in impute_dir.glob('*/*/Snakefile'):
            modify_snakefile_for_sbatch(path, project_name)
    else:
        raise ValueError(f'scheduler can only be qsub or sbatch, got {scheduler}')
    return


def calculate_3c_datasets(output_dir, fasta_path, cpu=10):
    output_dir = pathlib.Path(output_dir).absolute()
    project_name = output_dir.name
    scool_dir = output_dir / 'scool'
    snakemake_dir = scool_dir / 'snakemake'
    snakemake_dir.mkdir(exist_ok=True, parents=True)
    impute_dir = scool_dir / 'impute'
    impute_dir.mkdir(exist_ok=True)
    dataset_dir = scool_dir / 'dataset'
    dataset_dir.mkdir(exist_ok=True)

    # validate imputation
    none_success = []
    for sub_path in impute_dir.iterdir():
        if sub_path.is_dir():
            for chunk_dir in sub_path.glob('chunk*'):
                flag = False
                for path in chunk_dir.iterdir():
                    if path.name == 'Success':
                        flag = True
                if not flag:
                    none_success.append(chunk_dir)
    if len(none_success) != 0:
        print('These imputation directories did not successfully executed, check and run imputation again:')
        for path in none_success:
            print(path)
        raise FileNotFoundError

    # Compartment calling command
    compartment_input_dir = impute_dir / '100K'
    compartment_cell_table = pd.Series({
        path.name.split('.')[0]: str(path)
        for path in compartment_input_dir.glob('*/*.cool')
    })
    compartment_cell_table_path = compartment_input_dir / 'cell_table.tsv'
    compartment_cell_table.to_csv(compartment_cell_table_path, sep='\t', header=None)
    cpg_path = compartment_input_dir / 'cpg_ratio.hdf'
    # compute CpG ratio first
    cpg_ratio_cmd = f'hicluster cpg-ratio --cell_url {compartment_cell_table.iloc[0]} ' \
                    f'--fasta_path {fasta_path} --hdf_output_path {cpg_path}'
    execute_command(cpg_ratio_cmd)
    compartment_cmd = f'hicluster compartment ' \
                      f'--cell_table_path {compartment_cell_table_path} ' \
                      f'--output_prefix {dataset_dir / project_name} ' \
                      f'--cpg_profile_path {cpg_path} ' \
                      f'--cpu {cpu}'

    # Domain calling command
    domain_input_dir = impute_dir / '25K'
    domain_cell_table = pd.Series({
        path.name.split('.')[0]: str(path)
        for path in domain_input_dir.glob('*/*.cool')
    })
    domain_cell_table_path = domain_input_dir / 'cell_table.tsv'
    domain_cell_table.to_csv(domain_cell_table_path, sep='\t', header=None)
    domain_cmd = f'hicluster domain ' \
                 f'--cell_table_path {domain_cell_table_path} ' \
                 f'--output_prefix {dataset_dir / project_name} ' \
                 f'--resolution 25000 ' \
                 f'--window_size 10 ' \
                 f'--cpu {cpu}'

    # Embedding command
    embedding_dir = dataset_dir / 'embedding'
    embedding_dir.mkdir(exist_ok=True)
    embedding_cmd = f'hicluster embedding ' \
                    f'--cell_table_path {compartment_cell_table_path} ' \
                    f'--output_dir {embedding_dir} ' \
                    f'--dim 50 ' \
                    f'--dist 1000000 ' \
                    f'--resolution 100000 ' \
                    f'--scale_factor 100000 ' \
                    f'--norm_sig --save_raw ' \
                    f'--cpu {cpu}'

    # prepare qsub
    qsub_dir = snakemake_dir / 'qsub'
    qsub_dir.mkdir(exist_ok=True)
    with open(qsub_dir / 'dataset_commands.txt', 'w') as f:
        f.write('\n'.join([compartment_cmd, domain_cmd, embedding_cmd]))
    qsub_str = f'yap qsub --command_file_path {qsub_dir}/snakemake_cmd.txt ' \
               f'--working_dir {qsub_dir} --project_name y{project_name} ' \
               f'--total_cpu {int(cpu * 3)} --qsub_global_parms "-pe smp={cpu};-l h_vmem=5G"'
    with open(qsub_dir / 'qsub.sh', 'w') as f:
        f.write(qsub_str)
    print('Run this command to generate datasets')
    print(f"sh {qsub_dir / 'qsub.sh'}")
    return



import glob
import pathlib
import subprocess
import os
import pandas as pd

import cemba_data
from .m3c import m3c_config_str
from .mc import mc_config_str
from .mct import mct_config_str
from ...utilities import get_configuration

# Load defaults
PACKAGE_DIR = pathlib.Path(cemba_data.__path__[0])
INHOUSE_SERVERS = ['bpho', 'gale', 'cemba', 'oberon',
                   'login1.stampede2.tacc.utexas.edu',
                   'login2.stampede2.tacc.utexas.edu',
                   'login3.stampede2.tacc.utexas.edu',
                   'login4.stampede2.tacc.utexas.edu',
                   'login5.stampede2.tacc.utexas.edu',
                   'login6.stampede2.tacc.utexas.edu']


def prepare_uid_snakefile(uid_dir, config_str, snake_template):
    cell_ids = [path.name.split('.')[0][:-3] for path in (uid_dir / 'fastq').glob('*R1.fq.gz')]
    cell_id_str = f'CELL_IDS = {cell_ids}\n'

    # no file in this UID, do not make snakefile
    if len(cell_ids) == 0:
        return

    total_snakefile = config_str + cell_id_str + snake_template
    with open(uid_dir / 'Snakefile', 'w') as f:
        f.write(total_snakefile)
    return


def validate_mapping_config(output_dir):
    output_dir = pathlib.Path(output_dir).absolute()
    config = get_configuration(output_dir / 'mapping_config.ini')
    try:
        mode = config['mode']
    except KeyError:
        raise KeyError('mode not found in the config file.')

    if mode == 'mc':
        config_str = mc_config_str(config)
    elif mode == 'mct':
        config_str = mct_config_str(config)
    elif mode == 'm3c':
        config_str = m3c_config_str(config)
    else:
        raise ValueError(f'Unknown mode {mode}')

    print(f'Mapping config file looks good. Here is what will be used in generating Snakefile:\n{config_str}')
    return


def make_snakefile(output_dir):
    output_dir = pathlib.Path(output_dir).absolute()
    config = get_configuration(output_dir / 'mapping_config.ini')
    try:
        mode = config['mode']
    except KeyError:
        raise KeyError('mode not found in the config file.')

    if mode == 'mc':
        config_str = mc_config_str(config)
    elif mode == 'mct':
        config_str = mct_config_str(config)
    elif mode == 'm3c':
        config_str = m3c_config_str(config)
    else:
        raise ValueError(f'Unknown mode {mode}')
    print('Making Snakefile based on mapping config INI file. The parameters are:')
    print(config_str)

    with open(PACKAGE_DIR / f'mapping/Snakefile_template/{mode}.Snakefile') as f:
        snake_template = f.read()

    for sub_dir in output_dir.iterdir():
        if sub_dir.is_dir():
            if sub_dir.name not in ['stats', 'snakemake']:
                prepare_uid_snakefile(uid_dir=sub_dir,
                                      config_str=config_str,
                                      snake_template=snake_template)
    return


def write_qsub_commands(output_dir, cores_per_job, memory_gb_per_core, script_dir):
    memory_per_core = int(memory_gb_per_core[:-1]) * 1000
    cmds = {}
    snake_files = list(output_dir.glob('*/Snakefile'))
    for snake_file in snake_files:
        uid = snake_file.parent.name
        cmd = f'snakemake ' \
              f'-d {snake_file.parent} ' \
              f'--snakefile {snake_file} ' \
              f'-j {cores_per_job} ' \
              f'--default-resources mem_mb=100 ' \
              f'--resources mem_mb={int(cores_per_job * memory_per_core)} '
        cmds[uid] = cmd
    script_path = script_dir / 'snakemake_cmd.txt'
    with open(script_path, 'w') as f:
        try:
            uid_order = pd.read_csv(
                output_dir / 'stats/UIDTotalCellInputReadPairs.csv', index_col=0, squeeze=True, header=None
            ).sort_values(ascending=False)
            for uid in uid_order.index:
                if uid in cmds:
                    f.write(cmds.pop(uid) + '\n')
            try:
                assert len(cmds) == 0
            except AssertionError as e:
                print(cmds)
                print(uid_order)
                raise e
        except FileNotFoundError:
            # uid_order file do not exist (when starting from cell FASTQs)
            for cmd in cmds.values():
                f.write(cmd + '\n')
    return script_path


def write_sbatch_commands(output_dir, cores_per_job, script_dir, total_mem_mb, queue):
    output_dir_name = output_dir.name
    cmds = {}
    snake_files = list(output_dir.glob('*/Snakefile'))
    for snake_file in snake_files:
        uid = snake_file.parent.name
        cmd = f'snakemake ' \
              f'-d $SCRATCH/{output_dir_name}/{snake_file.parent.name} ' \
              f'--snakefile $SCRATCH/{output_dir_name}/{snake_file.parent.name}/Snakefile ' \
              f'-j {cores_per_job} ' \
              f'--default-resources mem_mb=100 ' \
              f'--resources mem_mb={total_mem_mb} ' \
              f'--rerun-incomplete ' \
              f'&& test -f "$SCRATCH/{output_dir_name}/{snake_file.parent.name}/MappingSummary.csv.gz"'
        cmds[uid] = cmd
    script_path = script_dir / f'snakemake_{queue}_cmd.txt'
    with open(script_path, 'w') as f:
        try:
            uid_order = pd.read_csv(
                output_dir / 'stats/UIDTotalCellInputReadPairs.csv', index_col=0, squeeze=True, header=None
            ).sort_values(ascending=False)
            for uid in uid_order.index:
                if uid in cmds:
                    f.write(cmds.pop(uid) + '\n')
            try:
                assert len(cmds) == 0
            except AssertionError as e:
                print(cmds)
                print(uid_order)
                raise e
        except FileNotFoundError:
            # uid_order file do not exist (when starting from cell FASTQs)
            for cmd in cmds.values():
                f.write(cmd + '\n')
    return f'$SCRATCH/{output_dir_name}/snakemake/sbatch/snakemake_{queue}_cmd.txt'


def prepare_qsub(name, snakemake_dir, total_jobs, cores_per_job, memory_gb_per_core):
    output_dir = snakemake_dir.parent
    qsub_dir = snakemake_dir / 'qsub'
    qsub_dir.mkdir(exist_ok=True)
    script_path = write_qsub_commands(output_dir, cores_per_job, memory_gb_per_core, script_dir=qsub_dir)
    qsub_str = f"""
#!/bin/bash
#$ -N yap{name}
#$ -V
#$ -l h_rt=99:99:99
#$ -l s_rt=99:99:99
#$ -wd {qsub_dir}
#$ -e {qsub_dir}/qsub.error.log
#$ -o {qsub_dir}/qsub.output.log
#$ -pe smp 1
#$ -l h_vmem=3G

yap qsub \
--command_file_path {script_path} \
--working_dir {qsub_dir} \
--project_name y{name} \
--total_cpu {int(cores_per_job * total_jobs)} \
--qsub_global_parms "-pe smp={cores_per_job};-l h_vmem={memory_gb_per_core}"
"""
    qsub_total_path = qsub_dir / 'qsub.sh'
    with open(qsub_total_path, 'w') as f:
        f.write(qsub_str)
    print('#' * 60)
    print(f"IF YOU USE QSUB ON GALE: ")
    print(f"All snakemake commands need to be executed "
          f"were included in {qsub_total_path}")
    print(f"You just need to qsub this script to "
          f"map the whole library in {output_dir}")
    print(f"You can also change the per job parameters in {script_path} "
          f"or change the global parameters in {qsub_total_path}")
    print(f"Read 'yap qsub -h' if you want to have more options about sbatch. "
          f"Alternatively, you can sbatch the commands in {script_path} by yourself, "
          f"as long as they all get successfully executed.")
    print('#' * 60 + '\n')
    return


def prepare_sbatch(name, snakemake_dir, queue):
    output_dir = snakemake_dir.parent
    output_dir_name = output_dir.name
    mode = get_configuration(output_dir / 'mapping_config.ini')['mode']

    if queue == 'skx-normal':
        sbatch_cores_per_job = 96
        if mode == 'm3c':
            time_str = "7:00:00"
            total_mem_mb = 160000
        elif mode == 'mc':
            time_str = "6:00:00"
            total_mem_mb = 192000
        elif mode == 'mct':
            time_str = "6:00:00"
            total_mem_mb = 192000
        else:
            raise KeyError(f'Unknown mode {mode}')
    elif queue == 'normal':
        sbatch_cores_per_job = 64
        if mode == 'm3c':
            time_str = "48:00:00"
            total_mem_mb = 90000
        elif mode == 'mc':
            time_str = "48:00:00"
            total_mem_mb = 112000
        elif mode == 'mct':
            time_str = "48:00:00"
            total_mem_mb = 112000
        else:
            raise KeyError(f'Unknown mode {mode}')
    else:
        raise ValueError(f'Unknown queue {queue}')
    sbatch_dir = snakemake_dir / 'sbatch'
    sbatch_dir.mkdir(exist_ok=True)

    script_path = write_sbatch_commands(output_dir,
                                        cores_per_job=sbatch_cores_per_job,
                                        script_dir=sbatch_dir,
                                        total_mem_mb=total_mem_mb,
                                        queue=queue)
    # the path here is using stampede path
    sbatch_cmd = f'yap sbatch ' \
                 f'--project_name {name}_{queue} ' \
                 f'--command_file_path {script_path} ' \
                 f'--working_dir $SCRATCH/{output_dir_name}/snakemake/sbatch ' \
                 f'--time_str {time_str} ' \
                 f'--queue {queue}'
    sbatch_total_path = sbatch_dir / f'sbatch-{queue}-queue.sh'
    with open(sbatch_total_path, 'w') as f:
        f.write(sbatch_cmd)

    print('#' * 40)
    print(f'For running jobs on the STAMPEDE2 {queue} queue:')
    print(f"All snakemake commands need to be executed "
          f"were included in {sbatch_total_path}")
    print(f"You just need to run this script to "
          f"map the whole library in {output_dir}. "
          f"Note that this script will not return until all the mapping job finished. "
          f"So you better run this script with nohup or in a screen.")
    print(f"You can also change "
          f"the per job parameters in {script_path} "
          f"or change the global parameters in {sbatch_total_path}")
    print(f"Read 'yap sbatch -h' if you want to have more options about sbatch. "
          f"Alternatively, you can sbatch the commands in "
          f"$SCRATCH/{output_dir_name}/snakemake/sbatch/sbatch.sh by yourself, "
          f"as long as they all get successfully executed.")
    print('#' * 40 + '\n')
    return


def prepare_run(output_dir, total_jobs=12, cores_per_job=10, memory_gb_per_core='5G', name=None):
    config = get_configuration(output_dir / 'mapping_config.ini')
    mode = config['mode']
    if mode in ['mc', 'm3c'] and cores_per_job < 4:
        raise ValueError(f'cores must >= 4 to run this pipeline.')
    elif mode == 'mct' and cores_per_job < 10:
        raise ValueError(f'cores must >= 10 to run this pipeline.')

    output_dir = pathlib.Path(output_dir).absolute()
    if name is None:
        name = output_dir.name
    snakemake_dir = output_dir / 'snakemake'
    snakemake_dir.mkdir(exist_ok=True)

    # this is only some automatic code for ecker lab...
    # so conditioned by the host name
    try:
        host_name = os.environ['HOSTNAME']
    except KeyError:
        host_name = subprocess.run('hostname', stdout=subprocess.PIPE, encoding='utf8').stdout
        if not isinstance(host_name, str):
            host_name = 'unknown'
    if any([host_name.startswith(s) for s in INHOUSE_SERVERS]):
        prepare_qsub(name=name,
                     snakemake_dir=snakemake_dir,
                     total_jobs=total_jobs,
                     cores_per_job=cores_per_job,
                     memory_gb_per_core=memory_gb_per_core)
        prepare_sbatch(name=name, snakemake_dir=snakemake_dir, queue='normal')
        prepare_sbatch(name=name, snakemake_dir=snakemake_dir, queue='skx-normal')
    else:
        script_path = write_qsub_commands(output_dir, cores_per_job, memory_gb_per_core, script_dir=snakemake_dir)
        print(f"All snakemake commands need to be executed were summarized in {script_path}")
        print(f"You need to execute them based on the computational environment you have "
              f"(e.g., use a job scheduler or run locally).")

    print(f"Once all commands are executed successfully, use 'yap summary' to generate final mapping summary.")
    return


def start_from_cell_fastq(output_dir, fastq_pattern, config_path):
    output_dir = pathlib.Path(output_dir).absolute()
    if output_dir.exists():
        raise FileExistsError(f'Output dir {output_dir} already exist, please delete it or use another path.')
    output_dir.mkdir()
    subprocess.run(['cp', config_path, f'{output_dir}/mapping_config.ini'], check=True)
    stats_dir = output_dir / 'stats'
    stats_dir.mkdir(exist_ok=True)

    # parse fastq patterns
    fastq_paths = [pathlib.Path(p).absolute() for p in glob.glob(fastq_pattern)]
    r1_records = {}
    r2_records = {}
    for path in fastq_paths:
        *cell_id, suffix = path.name.split('-')
        cell_id = '-'.join(cell_id)
        if suffix == 'R1.fq.gz':
            if cell_id in r1_records:
                raise ValueError(f'Found duplicated cell ID: {cell_id}, '
                                 f'File caused this error: {path}')
            r1_records[cell_id] = path
        elif suffix == 'R2.fq.gz':
            if cell_id in r2_records:
                raise ValueError(f'Found duplicated cell ID: {cell_id}, '
                                 f'File caused this error: {path}')
            r2_records[cell_id] = path
        else:
            raise ValueError(
                f'Unable to parse read type. Expect file name ends with "-R1.fq.gz" or "-R2.fq.gz", '
                f'File caused this error: {path}'
            )
    fastq_df = pd.DataFrame({'R1Path': r1_records, 'R2Path': r2_records})

    # make symlink of fastq files, using dir structure of demultiplex
    # the cells are randomly grouped though, max group is 64
    groups = min(64, fastq_df.shape[0])
    for i, (cell_id, (r1_path, r2_path)) in enumerate(fastq_df.sample(fastq_df.shape[0]).iterrows()):
        group_id = i % groups
        fastq_dir = output_dir / f'Group{group_id}/fastq'
        fastq_dir.mkdir(exist_ok=True, parents=True)

        # make symlinks
        new_r1_path = fastq_dir / r1_path.name
        new_r1_path.symlink_to(r1_path)
        new_r2_path = fastq_dir / r2_path.name
        new_r2_path.symlink_to(r2_path)

    # prepare scripts
    make_snakefile(output_dir)
    prepare_run(output_dir)
    return



import pandas as pd
import pathlib
import cemba_data

PACKAGE_DIR = pathlib.Path(cemba_data.__path__[0])


def prepare_mc_bulk(allc_table,
                    output_dir,
                    bigwig_contexts,
                    chrom_size_path,
                    extract_mcg=False,
                    bigwig_bin_size=50,
                    merge_allc_cpu=8,
                    mcg_context='CGN'):
    """
    Prepare the snakefile for merging single-cell ALLC files into pseudo-bulk

    Parameters
    ----------
    allc_table
        Path of the allc table. The allc table is a two column tsv file.
        The first columns is the absolute ALLC file paths;
        the second column is the group name of each file.
    output_dir
        Path of the output directory, will be created if not exist.
    extract_mcg
        Whether run the step to extract mCG sites from the merged ALLC.
        If your input ALLC only contains mCG sites, this can be skipped.
        Otherwise, this needs to be done before running the CG-DMR calling.
    bigwig_contexts
        mC contexts for generating the bigwig tracks
    chrom_size_path
        Path of the chromosome size file path
    bigwig_bin_size
        Bin size used to generate bigwig
    merge_allc_cpu
        Number of CPU to use in individual merge-allc job
    mcg_context
        mC context for extract_mcg step, only relevant when extract_mcg=True

    Returns
    -------

    """
    snakemake_template_path = PACKAGE_DIR / 'bulk/Snakefile_template/mc_bulk.Snakefile'
    output_dir = pathlib.Path(output_dir).absolute()
    output_dir.mkdir(exist_ok=True)

    # prepare ALLC path dict
    # allc_path to group
    allc_path = pd.read_csv(allc_table, sep='\t', index_col=0, header=None, squeeze=True)
    file_not_exist = allc_path[allc_path.index.map(lambda i: not pathlib.Path(i).exists())]
    if file_not_exist.size != 0:
        path_str = "\n".join(file_not_exist.index.tolist())
        raise FileNotFoundError(f'{file_not_exist.size} files do not exist:'
                                f'\n{path_str}')
    allc_dict = {group: paths.index.tolist() for group, paths in allc_path.groupby(allc_path)}

    # Prepare Snakefile
    allc_dir = output_dir / 'allc'
    allc_dir.mkdir(exist_ok=True)

    for group, paths in allc_dict.items():
        allc_list_path = allc_dir / f'{group}.allc_paths.txt'
        with open(allc_list_path, 'w') as f:
            f.write('\n'.join(paths))

    snakemake_parameters = f"""
merge_allc_cpu = {merge_allc_cpu}
extract_mcg = {extract_mcg}
mcg_context = '{mcg_context}'
bigwig_contexts = {bigwig_contexts}
bigwig_bin_size = {bigwig_bin_size}
chrom_size_path = '{chrom_size_path}'
groups = {list(allc_dict.keys())}

"""

    with open(snakemake_template_path) as f:
        snakemake_template = f.read()
    snakemake_str = snakemake_parameters + snakemake_template

    with open(output_dir / 'Snakefile', 'w') as f:
        f.write(snakemake_str)
    return



import shlex
import subprocess


def testing_cmd(command, expected_return_code=0):
    try:
        p = subprocess.run(shlex.split(command),
                           stderr=subprocess.PIPE,
                           stdout=subprocess.PIPE,
                           encoding='utf8',
                           check=True)
    except subprocess.CalledProcessError as e:
        if e.returncode == expected_return_code:
            return
        print(e.stderr)
        raise e
    return


COMMAND_TO_TEST = [
    'cutadapt --version',
    'bismark -version',
    'bowtie2 --version',
    'samtools --version',
    'tabix --version',
    'bgzip --version',
    'bedtools --version'
]


def testing_mapping_installation(mct=False):
    for command in COMMAND_TO_TEST:
        testing_cmd(command)

    # picard always return 1...
    testing_cmd('picard MarkDuplicates --version', 1)

    if mct:
        testing_cmd('STAR --version')

    # test ALLCools
    try:
        testing_cmd('allcools -h')
    except subprocess.CalledProcessError:
        print('"allcools -h" return error, see if allcools is installed. \n'
              'https://github.com/lhqing/ALLCools')
